//Fuzzy Game of Life
//Carlos Gershenson
//e-mail: carlos@jlagunez.iquimica.unam.mx

#include <graphics.h>
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include "tools.h"

typedef unsigned char uchar;
typedef unsigned int uint;

const MAXX=64;
const MAXY=32;

float m[MAXX][MAXY], n[MAXX+1][MAXY+1];
char g='',g2='';
int cyclic=1;       //flag for cyclic matrix
long gen;	    //# of generations
int s1,s2,s3,s4;    //stats
int pause=0,stats=0;//flags
char filename[12];

void FP(uint color, uchar r,uchar g,uchar b){
//change palete colors
//courtesy of Prof. Robert G. Murcio
     asm{
       MOV AH,10H     //{asigna int}
       MOV AL,10H     //{asigna subfucion}
       MOV BX,color   //{color a cambiar}
       MOV CH,g       //{cantidad de rojo, verde y azul}
       MOV CL,b
       MOV DH,r
       INT 10H
     }
}

int col(int x){
  switch (x){
    case 6: return 20;
    case 8: return 56;
    case 9: return 57;
    case 10: return 58;
    case 11: return 59;
    case 12: return 60;
    case 13: return 61;
    case 14: return 62;
    case 15: return 63;
    default: return x;
  }
}

void iniPalette(){
  outtextxy(0,0,"Loading Color Palette...");
  for (int i=0;i<16;i++)
    FP (col(i),i,0,(i*4)+3);
  for (i=0;i<=MAXX;i++) n[i][MAXY]=0;
  for (i=0;i<=MAXY;i++) n[MAXX][i]=0;
}

int loadFile(char * name){
  FILE * f;
  char cad[20000];  //yes, I know...
  char cx[3],cy[3],cv[10];
  int flag,x,y;
  float v;
  long pos,pos2;
  for (int i=0; i<MAXX; i++)
    for (int j=0; j<MAXY; j++)
      m[i][j]=0;
  if ((f = fopen(name, "r")) == NULL)
  {
     fprintf(stderr, "The file doesn't exist\n");
     getch();
     return(1);
  }
  fread(cad,20000, 1, f);
  pos=pos2=0;
  flag=1;
  while(flag){
    while((cad[pos]!='\n')&&(flag)){
      if ((cad[pos]=='E')||(pos>=20000)) flag=0;
      cx[pos-pos2]=cad[pos];
      pos++;
    }
    cx[pos-pos2]='\0';
    pos++;
    pos2=pos;
    while((cad[pos]!='\n')&&(flag)){
      if ((cad[pos]=='E')||(pos>=20000)) flag=0;
      cy[pos-pos2]=cad[pos];
      pos++;
    }
    cy[pos-pos2]='\0';
    pos++;
    pos2=pos;
    while((cad[pos]!='\n')&&(flag)){
      if ((cad[pos]=='E')||(pos>=20000)) flag=0;
      cv[pos-pos2]=cad[pos];
      pos++;
    }
    cv[pos-pos2]='\0';
    pos++;
    pos2=pos;
    if (flag){
      m[atoi(cx)][atoi(cy)]=atof(cv);
    }
    cout<<"x:"<<cx<<" y:"<<cy<<" v:"<<cv<<'\n';
  }
  fclose(f);
  return(0);
}

void paint (){
  BeforePaint();
  setcolor(15);
  line(0,320,640,320);
  outtextxy(0,330,"Fuzzy Game of Life");
  outtextxy(0,340,"Gershenson Technologies Inc., 1999");
  outtextxy(502,330,"(P)ause   (S)tats");
  outtextxy(550,340,"Esc to exit");

  for (int i=0; i<MAXX; i++)
    for (int j=0; j<MAXY; j++)
      if (m[i][j]>0){
	setcolor(m[i][j]*15);
	setfillstyle(1,m[i][j]*15);
	fillellipse((i*10)+5,(j*10)+5,4,4);
      }
  AfterPaint();
  if (pause){
    getch();
    pause=0;
  }
}

float around(int x, int y){//sum of cyclic neighbor cells of x,y
  int x1,x2,y1,y2;
  float v;
  if ((cyclic)&&((x-1)<0)) x1=MAXX-1; else if ((x-1)<0)x1=MAXX;else x1=x-1;
  if ((cyclic)&&((x+1)>=MAXX)) x2=0; else if ((x+1)>=MAXX)x2=MAXX;else x2=x+1;
  if ((cyclic)&&((y-1)<0)) y1=MAXY-1; else if ((y-1)<0)y1=MAXY;else y1=y-1;
  if ((cyclic)&&((y+1)>=MAXY)) y2=0; else if ((y+1)>=MAXY)y2=MAXY;else y2=y+1;
  v=n[x1][y1]+n[x1][y]+n[x1][y2]+n[x][y1]+
    n[x][y2]+n[x2][y1]+n[x2][y]+n[x2][y2];
  return v;
}

void trans(){
  float v,x;
  s1=s2=s3=s4=0;
  for (int i=0; i<MAXX; i++)
    for (int j=0; j<MAXY; j++)
      n[i][j]=m[i][j];
  for (i=0; i<MAXX; i++)
    for (int j=0; j<MAXY; j++){
      v=around(i,j);
      x=n[i][j];
      if (x>=0.5){
	if ((v<=1)||(v>=4)) m[i][j]=0;
	if ((1<v)&&(v<2)) m[i][j]=v-1;
	if ((2<=v)&&(v<=3)) m[i][j]=1;
	if ((3<v)&&(v<4)) m[i][j]=4-v;
      }
      else{
	if ((2<v)&&(v<=3)) m[i][j]=v-2;
	if ((3<v)&&(v<4)) m[i][j]=4-v;
      }
      if(m[i][j]==1)
	s1++;
      else
	if(m[i][j]==0)
	  s2++;
	else
	  if(m[i][j]>=0.5)
	    s3++;
	  else
	    s4++;
  }
/*-If the sum of the values of the 8 cells around x (let-s call it C) <= 1
or >= 4, and x>.5 then x=0.
-If 1<C<2 and x>.5 then x=C-1
-If 2<=C<=3 and x>.5 then x=1
-If 3<C<4 and x>.5 then x=4-C
(these last 4 ifs are the equivalents for the ifs of a cell with value 1.
The next 2 would be the equivalent for a cell with value of 0)

-If x<.5 and 2<C<=3 then x=C-2
-If x<.5 and 3<C<4 then x=4-C
  */
}

void rnd(){
  int pct=-1,pct1=-1;
  float t;
  cout<<"Around what percentage should be with cells? (0-100)\n";
  while ((pct<0)||(pct>100)) cin>>pct;
  cout<<"Around what percentage of it should be with 1's? (0-100)\n";
  while ((pct1<0)||(pct1>100))cin>>pct1;
  for (int i=0; i<MAXX; i++)
    for (int j=0; j<MAXY; j++)
      if (pct>random(100))
	if (pct1>random(100))
	  m[i][j]=1;
	else
	  m[i][j]=random(101)*0.01;
      else
	m[i][j]=0;
}

void stts(){
  char cad[10];
  BeforePaint();
  setcolor(15);
  outtextxy(0,0,"Fuzzy Life Stats");
  outtextxy(0,30,"Number of Generations:");
  itoa(gen, cad, 10);
  outtextxy(190,30,cad);
  outtextxy(0,50,"Number of cells=1:");
  itoa(s1, cad, 10);
  outtextxy(190,50,cad);
  outtextxy(0,70,"Number of cells=0:");
  itoa(s2, cad, 10);
  outtextxy(190,70,cad);
  outtextxy(0,90,"Number of 0.5<=cells<1:");
  itoa(s3, cad, 10);
  outtextxy(190,90,cad);
  outtextxy(0,110,"Number of 0<cells<0.5:");
  itoa(s4, cad, 10);
  outtextxy(190,110,cad);
  outtextxy(0,200,"Press any key...");
  AfterPaint();
  getch();
  stats=0;
}

int main(void){
   randomize();
   while (g!='0'){
     clrscr();
     textcolor(9);
     cout<<"Wellcome to Fuzzy Life!\n"<<'\n';
     cout<<"1) Load from file\n";
     cout<<"2) Random\n";
     cout<<"3) Cyclic";if (cyclic) cout<<" (On)\n";else cout<<" (Off)\n";
     cout<<"0) Exit\n";
     g=getch();
     if (g==27) exit(0);
     switch (g){
       case'1':cout<<"Please enter the file name\n";
	       cin>>filename;
	       if (loadFile(filename))
		 g='';
	       break;
       case'2':rnd();break;
       case'3':if (cyclic)
		 cyclic=0;
	       else
		 cyclic=1;
	       break;
     }
     if ((g=='1')||(g=='2')){
       IniGraphics();
       iniPalette();
       g2='';
       gen=0;
       while (g2!=27){
	 while (!kbhit()){
	   if (stats) stts();
	   paint();
	   trans();
	   gen++;
	 }
	 g2=getch();
	 switch (g2){
	   case 'p':pause=1;break;
	   case 's':stats=1;break;
	 }
       }
       closegraph();
     }
   }
   return 0;
}
